#include <stddef.h>
#include <string.h>
#include <stdint.h>
#include <stdalign.h>
#include <stdbool.h>
#include <ctype.h>
#include <assert.h>
#include <stdio.h>

#include "io.h"
#include "parser.h"

#define UPCAST(D,B,x) ((struct D*)((char*)x-offsetof(struct D,B)))
#define ARRAY_LENGTH(x) (sizeof(x)/sizeof(*x))


const struct category categories[] = {
  {
    "Zeichen",
    updateCategoryCharacter,
    sizeof(struct statics),
    alignof(struct statics)
  },{
    "Zeilenumbruch",
    updateCategoryNewline,
    sizeof(struct statics),
    alignof(struct statics)
  },{
    "Buchstaben",
    updateCategoryLetter,
    sizeof(struct statics),
    alignof(struct statics)
  },{
    "Wörter",
    updateCategoryWord,
    sizeof(struct wordStatics),
    alignof(struct wordStatics)
  },{
    "Sätze",
    updateCategorySentence,
    sizeof(struct sentenceStatics),
    alignof(struct sentenceStatics)
  }
};

enum category_change_state updateCategoryCharacter(
  struct statics_context* context,
  struct statics* stat,
  int c
){
  (void)context;
  (void)stat;
  return c != EOF;
}

enum category_change_state updateCategoryNewline(
  struct statics_context* context,
  struct statics* stat,
  int c
){
  (void)context;
  (void)stat;
  return c == '\n';
}

enum category_change_state updateCategoryLetter(
  struct statics_context* context,
  struct statics* stat,
  int c
){
  (void)context;
  (void)stat;
  return !!isalpha(c);
}

enum category_change_state updateCategoryWord(
  struct statics_context* context,
  struct statics* stat,
  int c
){
  (void)c;

  struct wordStatics* this = UPCAST( wordStatics, statics, stat );
  enum category_change_state result = CATEGORY_INACTIVE;

  if( context->statics[CATEGORY_LETTER]->lastResult == CATEGORY_MATCHED ){
    this->currentWordLength++;
    result = CATEGORY_ACTIVE;
  }else{
    if( this->currentWordLength > 0 )
      result = CATEGORY_MATCHED;
    this->currentWordLength = 0;
  }

  return result;
}

enum category_change_state updateCategorySentence(
  struct statics_context* context,
  struct statics* stat,
  int c
){
  (void)c;

  struct sentenceStatics* this = UPCAST( sentenceStatics, statics, stat );
  enum category_change_state result = CATEGORY_INACTIVE;

  if( context->statics[CATEGORY_LETTER]->lastResult == CATEGORY_MATCHED ){
    this->currentSentenceWordCount++;
  }else if( c == '.' || c == '!' || c == '?' || c == EOF ){
    if( this->currentSentenceWordCount > 0 )
      result = CATEGORY_MATCHED;
    this->currentSentenceWordCount = 0;
  }

  if( result == CATEGORY_INACTIVE && this->currentSentenceWordCount )
    result = CATEGORY_ACTIVE;

  return result;
}

struct staticsResult callWithStaticsContext( void(*callback)( struct statics_context*, void* ), void* param ){

  static_assert(
    CATEGORY_COUNT == ARRAY_LENGTH(categories),
    "struct categories and enum category_index must match, but they don't"
  );

  // Declarations //
  unsigned order[CATEGORY_COUNT];
  size_t offsets[CATEGORY_COUNT];
  size_t memory_size;
  // //

  // Determinate order of statics structs, struct with largest alignement first //
  for( unsigned i=0; i<CATEGORY_COUNT; i++ )
    order[i] = i;
  for( unsigned i=0; i<CATEGORY_COUNT-1; i++ )
  for( unsigned j=0; j<CATEGORY_COUNT-1-j; j++ )
  if( categories[order[j+1]].statics_align > categories[order[j]].statics_align ){
    unsigned tmp = order[j+1];
    order[j+1] = order[j];
    order[j] = tmp;
  }
  //  //

  // Calculating required memory size and offsets //
  // Starting with largest required alignement minus 1
  // because the start of a char array has no alignement requirements
  memory_size = categories[order[0]].statics_align-1;
  for( unsigned i=0; i<CATEGORY_COUNT; i++ ){
    // calculate alignement offset of current chunck, should be zero anyway
    const size_t alignement = categories[order[i]].statics_align;
    size_t offset = alignement - (memory_size % alignement);
    if( offset == alignement )
      offset = 0;
    // Calculate offset of next chunck
    memory_size += offset;
    offsets[i] = memory_size;
    // Calculate end of chunck, total memory required by all chuncks
    memory_size += categories[order[i]].statics_size;
  }
  //  //


  { // in new blockscope

    // More declarations //
    struct staticsResult result;
    struct statics_context context;

    // allocate memory and initialize it with zeros
    char memory[memory_size];
    // Fill memory with zeros
    memset( memory, 0, memory_size );

    // add alignement offset to memory
    const size_t alignement = categories[order[0]].statics_align;
    size_t alignement_offset = alignement - ( (uintptr_t)memory % alignement );
    if( alignement_offset == alignement )
      alignement_offset = 0;
    char* aligned_memory = (void*)(memory + alignement_offset);

    // initialize context struct
    for( unsigned i=0; i<CATEGORY_COUNT; i++ )
      context.statics[order[i]] = (void*)(aligned_memory + offsets[i]);

    // Call callback
    (*callback)( &context, param );

    // Copy results
    for( unsigned i=0; i<CATEGORY_COUNT; i++ )
      result.statics[i] = *context.statics[i];

    // return results
    return result;
  }
}

void defaultStaticsCalculator( struct statics_context* context, void* param ){
  struct unknown_io* io = param;
  int ch;

  do {

    ch = io->getch( io->io_context );

    for( unsigned i=0; i<CATEGORY_COUNT; i++ ){

      struct statics* statics = context->statics[i];
      enum category_change_state result = categories[i].updateCategory(
        context, statics, ch
      );

      statics->lastResult = result;

      if( result == CATEGORY_MATCHED )
        statics->count++;

    }

  } while( ch != EOF );

}
